# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import pytest
from marshmallow import ValidationError

from sherlockml.clients.user import (
    UserClient,
    User,
    UserSchema,
    AuthenticationResponse,
    AuthenticationResponseSchema,
)

from tests.clients.fixtures import PROFILE

USER_ID = uuid.uuid4()


def test_user_schema():
    data = UserSchema().load({"userId": str(USER_ID)})
    assert data == User(id=USER_ID)


@pytest.mark.parametrize(
    "data", [{}, {"id": str(USER_ID)}, {"userId": "not-a-uuid"}]
)
def test_user_schema_invalid(data):
    with pytest.raises(ValidationError):
        UserSchema().load(data)


def test_authentication_response_schema():
    data = AuthenticationResponseSchema().load(
        {"account": {"userId": str(USER_ID)}}
    )
    assert data == AuthenticationResponse(user=User(id=USER_ID))


@pytest.mark.parametrize(
    "data", [{}, {"user": {"id": str(USER_ID)}}, {"account": "not-an-account"}]
)
def test_authentication_response_schema_invalid(data):
    with pytest.raises(ValidationError):
        AuthenticationResponseSchema().load(data)


def test_user_client_authenticated_user_id(mocker):
    mocker.patch.object(
        UserClient,
        "_get",
        return_value=AuthenticationResponse(user=User(id=USER_ID)),
    )

    schema_mock = mocker.patch(
        "sherlockml.clients.user.AuthenticationResponseSchema"
    )

    client = UserClient(PROFILE)

    assert client.authenticated_user_id() == USER_ID

    UserClient._get.assert_called_once_with(
        "/authenticate", schema_mock.return_value
    )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
from datetime import datetime, timedelta

import pytz
import requests
from requests.auth import AuthBase


AccessToken = namedtuple("AccessToken", ["token", "expires_at"])


class AccessTokenClient(object):
    """Client for getting access tokens for accessing SherlockML services."""

    def __init__(self, hudson_url):
        self._session = requests.Session()
        self.hudson_url = hudson_url

    def get_access_token(self, client_id, client_secret):
        endpoint = "{}/access_token".format(self.hudson_url)
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "client_credentials",
        }
        response = self._session.post(endpoint, json=payload)
        response.raise_for_status()
        body = response.json()

        token = body["access_token"]
        now = datetime.now(tz=pytz.utc)
        expires_at = now + timedelta(seconds=body["expires_in"])

        return AccessToken(token, expires_at)


class SherlockMLAuth(AuthBase):
    """Requests auth implementation for accessing SherlockML services.

    Parameters
    ----------
    auth_service_url : str
        The URL of the SherlockML authentication service
    client_id : str
        The SherlockML client ID to use for authentication
    client_secret : str
        The client secret associated with the client ID

    To perform an authenticated request against a SherlockML service, first
    construct an instance of this class:

    >>> auth = SherlockMLAuth('https://hudson.example.sherlockml.net',
                              your_client_id, your_client_secret)

    then pass it as the ``auth`` argument when making a request with
    ``requests``:

    >>> import requests
    >>> requests.get('http://service.example.sherlockml.net', auth=auth)

    You can also set it as the ``auth`` attribute on a
    :class:`requests.Session`, so that subsequent requests will be
    authenticated automatically:

    >>> import requests
    >>> session = requests.Session()
    >>> session.auth = auth
    """

    def __init__(self, auth_service_url, client_id, client_secret):
        self.auth_service_url = auth_service_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = None

    def _get_token(self):
        client = AccessTokenClient(self.auth_service_url)
        return client.get_access_token(self.client_id, self.client_secret)

    def __call__(self, request):

        if self.access_token is None:
            self.access_token = self._get_token()
        if self.access_token.expires_at < datetime.now(tz=pytz.utc):
            self.access_token = self._get_token()

        header_content = "Bearer {}".format(self.access_token.token)
        request.headers["Authorization"] = header_content

        return request



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple

import pytest
from marshmallow import Schema, fields, post_load

from sherlockml.clients.base import (
    BaseClient,
    Unauthorized,
    NotFound,
    BadResponseStatus,
    InvalidResponse,
)
from tests.clients.fixtures import PROFILE

AUTHORIZATION_HEADER_VALUE = "Bearer mock-token"
AUTHORIZATION_HEADER = {"Authorization": AUTHORIZATION_HEADER_VALUE}

HUDSON_URL = "https://hudson.test.domain.com"

BAD_RESPONSE_STATUSES = [
    (401, Unauthorized),
    (404, NotFound),
    (400, BadResponseStatus),
    (500, BadResponseStatus),
]

HTTP_METHODS = ["GET", "POST", "PUT", "DELETE"]


@pytest.fixture
def patch_sherlockmlauth(mocker):
    def _add_auth_headers(request):
        request.headers["Authorization"] = AUTHORIZATION_HEADER_VALUE
        return request

    mock_auth = mocker.patch(
        "sherlockml.clients.base.SherlockMLAuth",
        return_value=_add_auth_headers,
    )

    yield

    mock_auth.assert_called_once_with(
        HUDSON_URL, PROFILE.client_id, PROFILE.client_secret
    )


DummyObject = namedtuple("DummyObject", ["foo"])


class DummySchema(Schema):
    foo = fields.String(required=True)

    @post_load
    def make_test_object(self, data):
        return DummyObject(**data)


class DummyClient(BaseClient):
    SERVICE_NAME = "test-service"


SERVICE_URL = "https://test-service.{}".format(PROFILE.domain)


def test_get(requests_mock, patch_sherlockmlauth):

    requests_mock.get(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        json={"foo": "bar"},
    )

    client = DummyClient(PROFILE)
    assert client._get("/test", DummySchema()) == DummyObject(foo="bar")


def test_post(requests_mock, patch_sherlockmlauth):

    mock = requests_mock.post(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        json={"foo": "bar"},
    )

    client = DummyClient(PROFILE)
    response = client._post("/test", DummySchema(), json={"test": "payload"})

    assert response == DummyObject(foo="bar")
    assert mock.last_request.json() == {"test": "payload"}


def test_put(requests_mock, patch_sherlockmlauth):

    mock = requests_mock.put(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        json={"foo": "bar"},
    )

    client = DummyClient(PROFILE)
    response = client._put("/test", DummySchema(), json={"test": "payload"})

    assert response == DummyObject(foo="bar")
    assert mock.last_request.json() == {"test": "payload"}


def test_delete(requests_mock, patch_sherlockmlauth):

    requests_mock.delete(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        json={"foo": "bar"},
    )

    client = DummyClient(PROFILE)
    response = client._delete("/test", DummySchema())

    assert response == DummyObject(foo="bar")


@pytest.mark.parametrize("status_code, exception", BAD_RESPONSE_STATUSES)
@pytest.mark.parametrize("http_method", HTTP_METHODS)
@pytest.mark.parametrize("check_status", [True, False])
def test_bad_responses(
    requests_mock,
    patch_sherlockmlauth,
    status_code,
    exception,
    http_method,
    check_status,
):

    mock_method = getattr(requests_mock, http_method.lower())
    mock_method(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        status_code=status_code,
        json={"foo": "bar"},
    )

    client = DummyClient(PROFILE)
    method = getattr(client, "_{}".format(http_method.lower()))
    if check_status:
        with pytest.raises(exception):
            method("/test", DummySchema(), check_status=check_status)
    else:
        method("/test", DummySchema(), check_status=check_status)


@pytest.mark.parametrize("status_code, exception", BAD_RESPONSE_STATUSES)
@pytest.mark.parametrize("http_method", HTTP_METHODS)
@pytest.mark.parametrize("check_status", [True, False])
def test_raw_bad_responses(
    requests_mock,
    patch_sherlockmlauth,
    status_code,
    exception,
    http_method,
    check_status,
):

    mock_method = getattr(requests_mock, http_method.lower())
    mock_method(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        status_code=status_code,
    )

    client = DummyClient(PROFILE)
    method = getattr(client, "_{}_raw".format(http_method.lower()))
    if check_status:
        with pytest.raises(exception):
            method("/test", check_status=check_status)
    else:
        method("/test", check_status=check_status)


@pytest.mark.parametrize("http_method", HTTP_METHODS)
def test_invalid_json(requests_mock, patch_sherlockmlauth, http_method):

    mock_method = getattr(requests_mock, http_method.lower())
    mock_method(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        text="invalid-json",
    )

    client = DummyClient(PROFILE)
    method = getattr(client, "_{}".format(http_method.lower()))
    with pytest.raises(InvalidResponse, match="not valid JSON"):
        method("/test", DummySchema())


@pytest.mark.parametrize("http_method", HTTP_METHODS)
def test_malformatted_json(requests_mock, patch_sherlockmlauth, http_method):

    mock_method = getattr(requests_mock, http_method.lower())
    mock_method(
        "{}/test".format(SERVICE_URL),
        request_headers=AUTHORIZATION_HEADER,
        json={"bad": "json"},
    )

    client = DummyClient(PROFILE)
    method = getattr(client, "_{}".format(http_method.lower()))
    with pytest.raises(InvalidResponse, match="not match expected format"):
        method("/test", DummySchema())



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
from enum import Enum

from marshmallow import Schema, fields, post_load
from marshmallow_enum import EnumField

from sherlockml.clients.base import BaseClient


class ServerStatus(Enum):
    CREATING = "creating"
    RUNNING = "running"
    ERROR = "error"
    DESTROYED = "destroyed"


Service = namedtuple("Service", ["name", "host", "port", "scheme", "uri"])


class ServiceSchema(Schema):

    name = fields.Str(required=True)
    host = fields.Str(required=True)
    port = fields.Int(required=True)
    scheme = fields.Str(required=True)
    uri = fields.Str(required=True)

    @post_load
    def make_service(self, data):
        return Service(**data)


Server = namedtuple(
    "Server",
    [
        "id",
        "project_id",
        "owner_id",
        "name",
        "type",
        "milli_cpus",
        "memory_mb",
        "created_at",
        "status",
        "services",
    ],
)


class ServerSchema(Schema):

    id = fields.UUID(data_key="instanceId", required=True)
    project_id = fields.UUID(data_key="projectId", required=True)
    owner_id = fields.UUID(data_key="ownerId", required=True)
    name = fields.Str(required=True)
    type = fields.Str(data_key="instanceType", required=True)
    milli_cpus = fields.Int(data_key="milliCpus", required=True)
    memory_mb = fields.Int(data_key="memoryMb", required=True)
    created_at = fields.DateTime(data_key="createdAt", required=True)
    status = EnumField(ServerStatus, by_value=True, required=True)
    services = fields.Nested(ServiceSchema, many=True, required=True)

    @post_load
    def make_server(self, data):
        return Server(**data)


class ServerIdSchema(Schema):

    instanceId = fields.UUID(required=True)

    @post_load
    def make_server_id(self, data):
        return data["instanceId"]


class ServerClient(BaseClient):

    SERVICE_NAME = "galleon"

    def create(
        self,
        project_id,
        server_type,
        milli_cpus,
        memory_mb,
        name=None,
        image_version=None,
        initial_environment_ids=None,
    ):

        payload = {
            "instanceType": server_type,
            "milliCpus": milli_cpus,
            "memoryMb": memory_mb,
        }
        if name:
            payload["name"] = name
        if image_version:
            payload["typeVersion"] = image_version
        if initial_environment_ids:
            payload["environmentIds"] = [
                str(env_id) for env_id in initial_environment_ids
            ]

        return self._post(
            "/instance/{}".format(project_id), ServerIdSchema(), json=payload
        )

    def get(self, project_id, server_id):
        endpoint = "/instance/{}/{}".format(project_id, server_id)
        return self._get(endpoint, ServerSchema())

    def list(self, project_id, name=None):
        endpoint = "/instance/{}".format(project_id)
        params = {"name": name} if name is not None else None
        return self._get(endpoint, ServerSchema(many=True), params=params)



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sherlockml.clients.user import UserClient
from sherlockml.clients.project import ProjectClient
from sherlockml.clients.server import ServerClient
from sherlockml.clients.cluster import ClusterClient
from sherlockml.clients.secret import SecretClient


CLIENT_FOR_RESOURCE = {
    "user": UserClient,
    "project": ProjectClient,
    "server": ServerClient,
    "cluster": ClusterClient,
    "secret": SecretClient,
}


def for_resource(resource):
    try:
        return CLIENT_FOR_RESOURCE[resource]
    except KeyError:
        raise ValueError(
            "unsupported resource {}, choose one of {}".format(
                resource, set(CLIENT_FOR_RESOURCE.keys())
            )
        )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime, timedelta

import pytest
import pytz

from sherlockml.clients.auth import (
    AccessToken,
    AccessTokenClient,
    SherlockMLAuth,
)


MOCK_HUDSON_URL = "https://hudson.example.com"
MOCK_CLIENT_ID = "client-id"
MOCK_CLIENT_SECRET = "client-secret"
MOCK_ACCESS_TOKEN_MATERIAL = "access-token"
NOW = datetime.now(tz=pytz.utc)


@pytest.fixture
def mock_datetime_now(mocker):
    datetime_mock = mocker.patch("sherlockml.clients.auth.datetime")
    datetime_mock.now.return_value = NOW
    return datetime_mock


def test_access_token_client(requests_mock, mock_datetime_now):

    requests_mock.post(
        "{}/access_token".format(MOCK_HUDSON_URL),
        json={"access_token": MOCK_ACCESS_TOKEN_MATERIAL, "expires_in": 600},
    )

    client = AccessTokenClient(MOCK_HUDSON_URL)
    access_token = client.get_access_token(MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)

    assert requests_mock.last_request.json() == {
        "client_id": MOCK_CLIENT_ID,
        "client_secret": MOCK_CLIENT_SECRET,
        "grant_type": "client_credentials",
    }
    assert access_token.token == MOCK_ACCESS_TOKEN_MATERIAL
    assert access_token.expires_at == NOW + timedelta(minutes=10)


@pytest.mark.parametrize(
    "cached_token",
    [None, AccessToken("old-token", NOW - timedelta(minutes=10))],
)
def test_sherlockml_auth(mocker, cached_token):

    mock_client = mocker.Mock()
    mock_client.get_access_token.return_value = AccessToken(
        MOCK_ACCESS_TOKEN_MATERIAL, NOW + timedelta(minutes=10)
    )
    client_patch = mocker.patch(
        "sherlockml.clients.auth.AccessTokenClient", return_value=mock_client
    )

    auth = SherlockMLAuth(MOCK_HUDSON_URL, MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)

    unauthenticated_request = mocker.Mock(headers={})
    request = auth(unauthenticated_request)

    assert request.headers["Authorization"] == "Bearer access-token"
    client_patch.assert_called_once_with(MOCK_HUDSON_URL)
    mock_client.get_access_token.assert_called_once_with(
        MOCK_CLIENT_ID, MOCK_CLIENT_SECRET
    )
    assert auth.access_token is not None


def test_sherlockml_auth_cached(mocker):

    mock_client = mocker.Mock()
    mocker.patch(
        "sherlockml.clients.auth.AccessTokenClient", return_value=mock_client
    )

    auth = SherlockMLAuth(MOCK_HUDSON_URL, MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)
    auth.access_token = AccessToken(
        MOCK_ACCESS_TOKEN_MATERIAL, NOW + timedelta(minutes=10)
    )

    unauthenticated_request = mocker.Mock(headers={})
    request = auth(unauthenticated_request)

    assert request.headers["Authorization"] == "Bearer access-token"
    mock_client.get_access_token.assert_not_called()



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import posixpath
from contextlib import contextmanager
import tempfile
import shutil
import uuid

import boto3
import pytest

from sherlockml.clients.secret import DatasetsSecrets

TEST_BUCKET_NAME = "sml-sfs-test"

TEST_FILE_NAME = "test_file"
TEST_FILE_CONTENT = b"this is a test file"
TEST_FILE_CONTENT_CHANGED = b"this is a different test file"
TEST_TREE = [
    "input/",
    "input/dir1/",
    "input/dir1/file1",
    "input/dir1/.file2",
    "input/.dir2/",
    "input/.dir2/file3",
    "output/",
    "output/file4",
    "empty/",
]
TEST_TREE_NO_HIDDEN_FILES = [
    "input/",
    "input/dir1/",
    "input/dir1/file1",
    "output/",
    "output/file4",
    "empty/",
]
VALID_ROOT_PATHS = ["", "./", "/"]
VALID_DIRECTORIES = [
    "input",
    "input/",
    "./input",
    "./input/",
    "/input",
    "/input/",
    "/input/dir1/",
    "/input/.dir2",
]
VALID_FILES = [
    "input/dir1/file1",
    "./input/dir1/file1",
    "/input/dir1/file1",
    "input/.dir2/file3",
]
EMPTY_DIRECTORY = "empty/"
INVALID_PATHS = [
    "inp",
    "inp/",
    "./inp",
    "./inp/",
    "/inp",
    "/inp/",
    "input/dir1/fil",
    "./input/dir1/fil",
    "/input/dir1/fil",
    "input/dir1/file1/",
    "./input/dir1/file1/",
    "/input/dir1/file1/",
]
TEST_DIRECTORY = "test_directory"
TEST_NON_EXISTENT = "test_non_existent"


@pytest.fixture
def project_env(monkeypatch):
    monkeypatch.setenv("SHERLOCKML_PROJECT_ID", str(uuid.uuid4()))


def _test_secrets():
    return DatasetsSecrets(
        bucket=TEST_BUCKET_NAME,
        access_key=os.environ["AWS_ACCESS_KEY_ID"],
        secret_key=os.environ["AWS_SECRET_ACCESS_KEY"],
        verified=True,
    )


@pytest.fixture
def mock_secret_client(mocker):
    mock_client = mocker.Mock()
    mock_client.datasets_secrets.return_value = _test_secrets()
    mocker.patch("sherlockml.client", return_value=mock_client)


@pytest.fixture
def project_directory(request, mock_secret_client, project_env):

    project_id = os.environ["SHERLOCKML_PROJECT_ID"]

    # Make the empty directory
    _make_file(request, "")
    yield

    # Tear down
    client = _s3_client()
    response = client.list_objects_v2(
        Bucket=TEST_BUCKET_NAME, Prefix=project_id
    )
    objects = [{"Key": obj["Key"]} for obj in response["Contents"]]
    client.delete_objects(Bucket=TEST_BUCKET_NAME, Delete={"Objects": objects})


@pytest.fixture
def remote_file(request, project_directory):
    _make_file(request, TEST_FILE_NAME, TEST_FILE_CONTENT)
    return TEST_FILE_NAME


@pytest.fixture
def remote_tree(request, project_directory):
    for path in TEST_TREE:
        if path.endswith("/"):
            # Just touch, is a directory
            _make_file(request, path)
        else:
            # Add test file content
            _make_file(request, path, TEST_FILE_CONTENT)
    return TEST_TREE


_S3_CLIENT_CACHE = None


def _s3_client():
    global _S3_CLIENT_CACHE
    if _S3_CLIENT_CACHE is None:
        secrets = _test_secrets()
        boto_session = boto3.session.Session(
            aws_access_key_id=secrets.access_key,
            aws_secret_access_key=secrets.secret_key,
            region_name="eu-west-1",
        )
        _S3_CLIENT_CACHE = boto_session.client("s3")
    return _S3_CLIENT_CACHE


def _path_in_bucket(path):
    project_id = os.environ["SHERLOCKML_PROJECT_ID"]
    combined = posixpath.join(project_id, path.lstrip("/"))
    if path.endswith("/"):
        return posixpath.normpath(combined) + "/"
    else:
        return posixpath.normpath(combined)


def write_remote_object(path, content):
    client = _s3_client()
    client.put_object(
        Bucket=TEST_BUCKET_NAME,
        Key=_path_in_bucket(path),
        Body=content,
        ServerSideEncryption="AES256",
    )


def read_remote_object(path):
    client = _s3_client()
    obj = client.get_object(Bucket=TEST_BUCKET_NAME, Key=_path_in_bucket(path))
    return obj["Body"].read()


def _make_file(request, path, content=""):
    write_remote_object(path, content)

    def tear_down():
        _s3_client().delete_object(
            Bucket=TEST_BUCKET_NAME, Key=_path_in_bucket(path)
        )

    request.addfinalizer(tear_down)


@pytest.fixture
def local_file():
    with tempfile.NamedTemporaryFile(mode="wb") as f:
        f.write(TEST_FILE_CONTENT)
        f.flush()
        yield f.name


@pytest.fixture
def local_file_changed():
    with tempfile.NamedTemporaryFile(mode="wb") as f:
        f.write(TEST_FILE_CONTENT_CHANGED)
        f.flush()
        yield f.name


@contextmanager
def temporary_directory():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


@pytest.fixture
def local_tree():

    with temporary_directory() as temp_dir:

        for tree_path in TEST_TREE:

            path = os.path.join(temp_dir, tree_path)

            if tree_path.endswith("/"):
                if not os.path.exists(path):
                    os.makedirs(path)
            else:
                dirname = os.path.dirname(path)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                with open(path, "wb") as fp:
                    fp.write(TEST_FILE_CONTENT)

        yield temp_dir



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from marshmallow import ValidationError

from sherlockml.clients.secret import DatasetsSecrets, DatasetsSecretsSchema


TEST_SECRETS = DatasetsSecrets(
    bucket="test-bucket",
    access_key="test-access-key",
    secret_key="test-secret-key",
    verified=True,
)

TEST_SECRETS_BODY = {
    "bucket": TEST_SECRETS.bucket,
    "access_key": TEST_SECRETS.access_key,
    "secret_key": TEST_SECRETS.secret_key,
    "verified": TEST_SECRETS.verified,
}


def test_datasets_secrets_schema():
    assert DatasetsSecretsSchema().load(TEST_SECRETS_BODY) == TEST_SECRETS


def test_datasets_secrets_invalid():
    with pytest.raises(ValidationError):
        DatasetsSecretsSchema().load({})



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

import sherlockml.clients
from sherlockml.clients.user import UserClient


def test_for_resource():
    assert sherlockml.clients.for_resource("user") is UserClient


def test_for_resource_missing():
    with pytest.raises(ValueError):
        sherlockml.clients.for_resource("missing")



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid

import pytest
from marshmallow import ValidationError

from sherlockml.clients.project import ProjectSchema, Project, ProjectClient
from tests.clients.fixtures import PROFILE


USER_ID = uuid.uuid4()

PROJECT = Project(id=uuid.uuid4(), name="test-project", owner_id=uuid.uuid4())

PROJECT_BODY = {
    "projectId": str(PROJECT.id),
    "name": PROJECT.name,
    "ownerId": str(PROJECT.owner_id),
}


def test_project_schema():
    data = ProjectSchema().load(PROJECT_BODY)
    assert data == PROJECT


def test_project_schema_load_invalid():
    with pytest.raises(ValidationError):
        ProjectSchema().load({})


def test_project_client_list_accessible_by_user(mocker):
    mocker.patch.object(ProjectClient, "_get", return_value=[PROJECT])
    schema_mock = mocker.patch("sherlockml.clients.project.ProjectSchema")

    client = ProjectClient(PROFILE)
    assert client.list_accessible_by_user(USER_ID) == [PROJECT]

    schema_mock.assert_called_once_with(many=True)
    ProjectClient._get.assert_called_once_with(
        "/user/{}".format(USER_ID), schema_mock.return_value
    )


def test_project_client_get(mocker):
    mocker.patch.object(ProjectClient, "_get", return_value=PROJECT)
    schema_mock = mocker.patch("sherlockml.clients.project.ProjectSchema")

    client = ProjectClient(PROFILE)
    assert client.get(PROJECT.id) == PROJECT

    schema_mock.assert_called_once_with()
    ProjectClient._get.assert_called_once_with(
        "/project/{}".format(PROJECT.id), schema_mock.return_value
    )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from sherlockml.datasets import path


@pytest.mark.parametrize(
    "input_path, rationalised_path",
    [
        ("", "/"),
        ("./", "/"),
        ("/", "/"),
        ("path", "/path"),
        ("./path", "/path"),
        ("/path", "/path"),
        ("path/", "/path/"),
        ("./path/", "/path/"),
        ("/path/", "/path/"),
    ],
)
def test_rationalise_projectpath(input_path, rationalised_path):
    assert path.rationalise_projectpath(input_path) == rationalised_path


@pytest.mark.parametrize(
    "project_path, project_id, bucket_path",
    [
        ("", "id", "id/"),
        ("./", "id", "id/"),
        ("/", "id", "id/"),
        ("path", "id", "id/path"),
        ("./path", "id", "id/path"),
        ("/path", "id", "id/path"),
        ("path/", "id", "id/path/"),
        ("./path/", "id", "id/path/"),
        ("/path/", "id", "id/path/"),
    ],
)
def test_projectpath_to_bucketpath(project_path, project_id, bucket_path):
    result = path.projectpath_to_bucketpath(project_path, project_id)
    assert result == bucket_path


@pytest.mark.parametrize(
    "bucket_path, project_path",
    [
        ("id/", "/"),
        ("id/path", "/path"),
        ("id/path/", "/path/"),
        ("id/nested/path", "/nested/path"),
    ],
)
def test_bucketpath_to_projectpath(bucket_path, project_path):
    result = path.bucketpath_to_projectpath(bucket_path)
    assert result == project_path


@pytest.mark.parametrize(
    "project_path",
    [
        "input/path/to/somefile.csv",
        "./input/path/to/somefile.csv",
        "/input/path/to/somefile.csv",
        "input/path/to/somedir/",
        "./input/path/to/somedir/",
        "/input/path/to/somedir/",
    ],
)
def test_project_parent_directories(project_path):
    correct = ["/", "/input/", "/input/path/", "/input/path/to/"]
    result = path.project_parent_directories(project_path)
    assert set(result) == set(correct)
